#!/usr/bin/env python
# -*- coding:utf-8 -*-
from abc import ABC, abstractmethod


class Validator(ABC):
    """
    Custom validators need to inherit from Validator
    and must supply a validate() method to test various restrictions as needed.
    """

    def __set_name__(self, owner, name):
        self.private_name = '_' + name

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.private_name)

    def __set__(self, obj, value):
        self.validate(value)
        setattr(obj, self.private_name, value)

    @abstractmethod
    def validate(self, value):
        pass


"""
Demo validators.
1.OneOf: verifies that a value is one of a restricted set of options.
2.Number: verifies that a value is either an int or float. Optionally, 
  it verifies that a value is between a given minimum or maximum.
3.String: verifies that a value is a str. Optionally, it validates a given minimum or maximum length. 
  It can validate a user-defined predicate as well.
4.ClassType: verifies that a value's type in types.
"""


class OneOf(Validator):
    """
    verifies that a value is one of a restricted set of options.
    """

    def __init__(self, *options):
        self.options = set(options)

    def validate(self, value):
        if value not in self.options:
            raise ValueError(f'Expected {value!r} to be one of {self.options!r}')


class Number(Validator):
    """
    verifies that a value is either an int or float. Optionally,
    it verifies that a value is between a given minimum or maximum.
    """

    def __init__(self, minvalue=None, maxvalue=None):
        self.minvalue = minvalue
        self.maxvalue = maxvalue

    def validate(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError(f'Expected {value!r} to be an int or float')
        if self.minvalue is not None and value < self.minvalue:
            raise ValueError(
                f'Expected {value!r} to be at least {self.minvalue!r}'
            )
        if self.maxvalue is not None and value > self.maxvalue:
            raise ValueError(
                f'Expected {value!r} to be no more than {self.maxvalue!r}'
            )


class String(Validator):
    """
    verifies that a value is a str. Optionally,
    it validates a given minimum or maximum length. It can validate a user-defined predicate as well.
    """

    def __init__(self, minsize=None, maxsize=None, predicate=None):
        self.minsize = minsize
        self.maxsize = maxsize
        self.predicate = predicate

    def validate(self, value):
        if not isinstance(value, str):
            raise TypeError(f'Expected {value!r} to be an str')
        if self.minsize is not None and len(value) < self.minsize:
            raise ValueError(
                f'Expected {value!r} to be no smaller than {self.minsize!r}'
            )
        if self.maxsize is not None and len(value) > self.maxsize:
            raise ValueError(
                f'Expected {value!r} to be no bigger than {self.maxsize!r}'
            )
        if self.predicate is not None and not self.predicate(value):
            raise ValueError(
                f'Expected {self.predicate} to be true for {value!r}'
            )


class ClassType(Validator):
    """
    verifies that a value type in types.
    """

    def __init__(self, *types: type):
        self.__types: tuple[type] = types
        self.__type_names: list[str] = []
        for t in types:
            if not isinstance(t, type):
                raise TypeError(f"Excepted 'type' object, got a '{t}'")
            self.__type_names.append(t.__name__)

    def validate(self, value):
        value_type = type(value)
        if not issubclass(value_type, self.__types):
            raise TypeError(f"Expected \"{self.__type_names}\", got a '{value_type.__name__}'")


__all__ = ["Validator", "OneOf", "Number", "String", "ClassType"]
