#!/usr/bin/env python
# -*- coding:utf-8 -*-

from functools import wraps
from typing import Callable, Type, Any, Tuple, Dict, Union, List

from ._hook import _get_chain, _run_hook_func
from ._process import _do


def capture(exception: Union[Type[Exception], Tuple[Type[Exception]]] = BaseException, post: Callable = None):
    """
    Catch exceptions to decorated functions, including exceptions generated by other decorators.

    callback functions can and only support communication via the chain keyword parameter. example: callback() is ok,
    callback(chain=None) is ok, callback(chain=None, other=None) is ok(other arg will not be assigned), callback(other, chain=None) will happend exception

    :param post while callback function when happen exception
    :param exception: The base type exception to the exception that needs to be caught. After capture,
                    the exception will be "eaten" directly, and exception outside this range will be thrown.
    :return:
    """
    def _inner(func):
        @wraps(func)
        def _wrapper(*args, **kwargs):
            opts = {"exception": exception, "post": post, "args": args, "kwargs": kwargs, "stacklevel": 7}
            return _do(func=func, decorator_name=capture.__name__, args=args, kwargs=kwargs, opts=opts)

        return _wrapper

    return _inner


def __do_capture(func: Callable, args: Tuple = None, kwargs: Dict = None, opts: Dict = None) -> Any:
    exception: Union[Type[Exception], Tuple[Type[Exception]]] = opts.get("exception")
    post: Callable = opts.get("post")
    chain, func_new_kwargs = _get_chain(func, args, kwargs)
    try:
        return func(*args, **kwargs)
    except BaseException as e:
        if issubclass(type(exception), List):
            for e_ in exception:
                if not issubclass(type(e), e_):
                    raise
        else:
            if not issubclass(type(e), exception):
                raise
    finally:
        if post:
            _run_hook_func([post], func_new_kwargs)


__all__ = [capture]
