#!/usr/bin/env python
# -*- coding:utf-8 -*-
from functools import wraps
from time import sleep
from typing import Type, Callable, Dict, Any, Tuple, Union, List

from ._process import _do
from .._internal import _T
from .._internal._method_helper import func_full_args
from ..decorators._hook import _run_hook_func
from ..log import LoggerFactory

__logger = LoggerFactory.get_logger("retry")


def retry(frequency: int = 1, interval: int = 1, increasing: int = 0, check: Callable[[_T], bool] = lambda r: True,
          ignored_exception: Union[Type[Exception], Tuple[Type[Exception]]] = None,
          post: Callable[[_T, _T], None] = None):
    """
    Provides retry functionality, if function run success, will run once.
    unit second

    callback functions can and only support communication via the chain keyword parameter. example: callback() is ok,
    callback(chain=None) is ok, callback(chain=None, other=None) is ok(other arg will not be assigned),
    callback(other, chain=None) will happened exception

    :param check: Check whether the result is as expected, return the result if it is met, otherwise try again.
    :param ignored_exception: If it is the exception and its child exceptions, no retry is made
    :param post: If the original function is not executed.Two parameters are required, func_params and func_return
    the callback function is executed and the result of the callback function is returned,
    arguments to the original function are passed to the callback function
    :param frequency: number of executions.
    :param interval: retry interval, unit seconds.
    :param increasing: The incrementing interval

    Example:
        def post_hook(*args, **kwargs):
            print(args, kwargs)

        @retry(post=post_hook)
        def foo():

    """
    def _inner(func):
        @wraps(func)
        def _wrapper(*args, **kwargs):
            opts = {"frequency": frequency, "interval": interval, "increasing": increasing, "check": check,
                    "ignored_exception": ignored_exception, "post": post, "stacklevel": 7}
            return _do(func=func, decorator_name=retry.__name__,  args=args, kwargs=kwargs, opts=opts)

        return _wrapper

    return _inner


def retrying(func: Callable, args: Tuple = None, kwargs: Dict = None,
             frequency: int = 1, interval: int = 1, increasing: int = 0,
             check: Callable[[_T], bool] = lambda r: r,
             ignore_exception: Union[Type[Exception], Tuple[Type[Exception]]] = None,
             post: Callable[[_T, _T], None] = None) -> Any:
    """
    if function run success, will run once.
    :param func: The function to be retried.
    :param args: The parameters of the function to be retried must be provided as args.
    :param kwargs: The parameters of the function to be retried must be provided as kwargs.
    :param check: Check whether the result is as expected, return the result if it is met, otherwise try again.
    :param ignore_exception: If it is the exception and its child exceptions, no retry is made
    :param post: If the original function is not executed.Two parameters are required, func_params and func_return
    the callback function is executed and the result of the callback function is returned,
    arguments to the original function are passed to the callback function
    :param frequency: number of executions.
    :param interval: retry interval.
    :param increasing: The incrementing interval
    :return:
    """
    opts = {"frequency": frequency, "interval": interval, "increasing": increasing, "check": check,
            "ignore_exception": ignore_exception, "post": post, "stacklevel": 4}
    args_ = args if args else []
    kwargs_ = kwargs if kwargs else {}
    return __do_retry(func, args=args_, kwargs=kwargs_, opts=opts)


def repeat(frequency: int = 1, interval: int = 1, increasing: int = 0,
           ignored_exception: Union[Type[Exception], Tuple[Type[Exception]]] = None):
    """
    Repeat the function, return last result.
    :param frequency: number of executions
    :param interval: the time between executions, unit seconds
    :param increasing: the incrementing interval
    :param ignored_exception:
                if happened exception and exception not included in ignored_exception will interrupt execution
    :return: last execute result
    """
    def _inner(func):
        @wraps(func)
        def _wrapper(*args, **kwargs):
            opts = {"frequency": frequency, "interval": interval, "increasing": increasing,
                    "ignore_exception": ignored_exception, "stacklevel": 7}
            return _do(func=func, decorator_name=repeat.__name__, args=args, kwargs=kwargs, opts=opts)

        return _wrapper

    return _inner


def repeating(func, frequency: int = 1, interval: int = 1, increasing: int = 0,
              ignored_exception: Union[Type[Exception], Tuple[Type[Exception]]] = None,
              args: Tuple = None, kwargs: Dict = None) -> Any:
    """
    Repeat the function, return last result
    :param func: the function to execute
    :param frequency: number of executions
    :param interval: the time between executions, unit seconds
    :param increasing: the incrementing interval
    :param ignored_exception:
                    if happened exception and exception not included in ignored_exception will interrupt execution
    :param args: func needs args
    :param kwargs: func needs kwargs
    :return: last execute result
    """
    opts = {"frequency": frequency, "interval": interval, "increasing": increasing,
            "ignore_exception": ignored_exception, "stacklevel": 4}
    args_ = args if args else []
    kwargs_ = kwargs if kwargs else {}
    return __do_repeat(func, args=args_, kwargs=kwargs_, opts=opts)


def __do_retry(func: Callable = None, args: Tuple = None, kwargs: Dict = None, opts: Dict = None):
    interval: int = opts.get("interval")
    frequency: int = opts.get("frequency")
    increasing: int = opts.get("increasing")
    ignore_exception: Union[Type[Exception], Tuple[Type[Exception]]] = opts.get("ignore_exception")
    post: Callable[[_T, _T], None] = opts.get("post")
    check: Callable[[_T], bool] = opts.get("check")
    call_back_func_kwargs = func_full_args(func, args, kwargs)
    args_ = () if args is None else args
    kwargs_ = {} if kwargs is None else kwargs
    for _ in range(1, frequency + 1):
        # noinspection PyBroadException
        try:
            result = func(*args_, **kwargs_)
            if check(result):
                return result
            else:
                __logger.log(level=30, msg=f"check result fail!!!", stacklevel=4)
            __logger.log(level=20, msg=f"run function '{func.__name__}' "
                                       f"fail: retrying {_} time(s)", stacklevel=4)
            sleep(interval)
            interval += increasing
        except BaseException as e:
            if issubclass(type(ignore_exception), List):
                for e_ in ignore_exception:
                    if not issubclass(type(e), e_):
                        raise
            else:
                if not issubclass(type(e), ignore_exception):
                    raise
            __logger.log(level=20, msg=f"run function '{func.__name__}' exception {type(e).__name__}: {str(e)}."
                                       f" retrying {_} time(s)", stacklevel=4)
            sleep(interval)
            interval += increasing
    else:
        if post:
            post(*args_, **kwargs_)
            _run_hook_func(post, call_back_func_kwargs)


def __do_repeat(func, args: Tuple = None, kwargs: Dict = None, opts: Dict = None) -> Any:
    stacklevel: int = opts.get("stacklevel")
    interval: int = opts.get("interval")
    frequency: int = opts.get("frequency")
    increasing: int = opts.get("increasing")
    ignored_exception: Union[Type[Exception], Tuple[Type[Exception]]] = opts.get("ignored_exception")
    args_ = () if args is None else args
    kwargs_ = {} if kwargs is None else kwargs
    result = None
    for _ in range(1, frequency + 1):
        # noinspection PyBroadException
        try:
            result = func(*args_, **kwargs_)
        except BaseException as e:
            __logger.log(level=40, msg=f"run '{func.__name__}(args={args}, kwargs={kwargs})' happened exception, "
                                       f"result={result}, exception={str(e)}, run {_} time(s).", stacklevel=stacklevel)
            if ignored_exception and not issubclass(e.__class__, ignored_exception):
                raise
        finally:
            __logger.log(level=20, msg=f"run '{func.__name__}(args={args}, kwargs={kwargs})', "
                                       f"result={result}, run {_} time(s).", stacklevel=stacklevel)
        if _ == frequency:
            break
        sleep(interval)
        interval += increasing
    return result


__all__ = [retry, retrying, repeat, repeating]
