#!/usr/bin/env python
# -*- coding:utf-8 -*-
from abc import ABC, abstractmethod
from datetime import datetime
from functools import wraps
from types import NoneType
from typing import List, Any, Callable, Tuple, Dict, Union, Type

from ._process import _do
from .._internal._method_helper import func_full_args
from ..classes import ForceType


class Validator(ABC):

    def __set_name__(self, owner, name):
        self.private_name = '_' + name

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.private_name)

    def __set__(self, obj, value):
        self.validate(value)
        setattr(obj, self.private_name, value)

    @abstractmethod
    def validate(self, value):
        pass


class Valid(object):
    name = ForceType(str)
    datetime_format = ForceType(str, NoneType)
    empty = ForceType(bool)
    notEmpty = ForceType(bool)
    types = ForceType(tuple, NoneType)

    def __init__(self, name: str, length: int = None,
                 contain: Union[Any, NoneType] = None,
                 contained: Union[Any, NoneType] = None,
                 eq: Union[Any, NoneType] = None, le: Union[Any, NoneType] = None,
                 lt: Union[Any, NoneType] = None, ne: Union[Any, NoneType] = None,
                 ge: Union[Any, NoneType] = None, gt: Union[Any, NoneType] = None,
                 datetime_format: Union[str, NoneType] = None, empty: bool = False, notEmpty: bool = False,
                 types: [Tuple[Type] or NoneType] = None):
        """
        A metadata object that validates the validity of the parameter.
        :param name: Parameter name.
        :param length: Verify that the length of the parameter is the same as expected.
        :param contain: Verify that the parameter contains this option.
        :param contained: Verify that the option (iterable) contains parameters.
        :param eq: Verify that the parameters and options are equal.
        :param le: Whether the check parameter is less than or equal to the check value.
        :param lt: Whether the check parameter is less than the check value.
        :param ne: Whether the check parameter is not equal to the check value.
        :param ge: Check whether the parameter is greater than the check value.
        :param gt: Whether the check parameter is greater than or equal to the check value.
        :param datetime_format: If the parameter is time, verifies that the time is in the specified format.
        :param empty: Verify whether the parameter is empty
        :param notEmpty: The validation parameter is not empty
        """
        self.name: str = name
        self.length: int = length
        self.contain: Union[Any, NoneType] = contain
        self.contained: Union[Any, NoneType] = contained
        self.eq: Union[Any, NoneType] = eq
        self.le: Union[Any, NoneType] = le
        self.lt: Union[Any, NoneType] = lt
        self.ne: Union[Any, NoneType] = ne
        self.ge: Union[Any, NoneType] = ge
        self.gt: Union[Any, NoneType] = gt
        self.datetime_format: str = datetime_format
        self.empty: bool = empty
        self.notEmpty: bool = notEmpty
        self.types: [Tuple[Type] or NoneType] = types
        self.__types_names: List[str] = [t.__name__ for t in types]

    @property
    def type_names(self) -> List[str]:
        return self.__types_names


def validate(conditions: List[Valid]):
    """
    Parameters to the validation method/function.
    :param conditions: Validation conditions, a collection of instances of Valid
    """

    def __inner(func):
        @wraps(func)
        def __wrapper(*args, **kwargs):
            opts = {"conditions": conditions, "stacklevel": 7}
            return _do(func=func, decorator_name=validate.__name__, args=args, kwargs=kwargs, opts=opts)

        return __wrapper

    return __inner


def __do_validate(func: Callable, args: Tuple = None, kwargs: Dict = None, opts: Dict = None):
    conditions: List[Valid] = opts.get("conditions", [])
    args_ = () if args is None else args
    kwargs_ = {} if kwargs is None else kwargs
    func_parameters = func_full_args(func, args_, kwargs_)
    for valid in conditions:
        if not isinstance(valid, Valid):
            raise TypeError(f"valid check metadata type error , excepted type '{Valid.__name__}', "
                            f"got a '{type(valid).__name__}'")
        parameter = func_parameters.get(valid.name)
        if valid.empty:
            assert not parameter, f"valid check parameter is empty, but '{valid.name}=({parameter})' is not empty."
        if valid.notEmpty:
            assert parameter, f"valid check parameter is not empty, but '{valid.name}=({parameter})' is empty."
        if valid.types is not None:
            type_ = type(parameter)
            assert issubclass(type_, valid.types), f"valid check parameter type, expected '{valid.type_names}', " \
                                                   f"got '{type_.__name__}'"
        if valid.length is not None:
            assert len(parameter) == valid.length, f"Valid length check: " \
                                                   f"expected '{valid.length}', got a '{len(parameter)}'."
        if valid.contain is not None:
            assert valid.contain in parameter, f"Valid contain check: " \
                                               f"expected '{parameter}' contain '{valid.name}=({valid.contain})', " \
                                               f"got not contain."
        if valid.contained is not None:
            assert parameter in valid.contained, f"Valid contained check: " \
                                                 f"expected '{valid.name}=({valid.contain})' contain '{parameter}', " \
                                                 f"got not contain."
        if valid.eq is not None:
            assert parameter == valid.eq, f"Valid eq check: expected '{parameter}' = '{valid.name}=({valid.eq})', " \
                                          f"got '{parameter}' != '{valid.name}=({valid.eq})'."
        if valid.lt is not None:
            assert parameter < valid.lt, f"Valid lt check: expected '{parameter}' < '{valid.name}=({valid.lt})', " \
                                         f"got  '{parameter}' >= '{valid.name}=({valid.lt})'."
        if valid.le is not None:
            assert parameter <= valid.le, f"Valid le check: expected '{parameter}' <= '{valid.name}=({valid.le})', " \
                                          f"got  '{parameter}' > '{valid.name}=({valid.le})'."
        if valid.ne is not None:
            assert parameter != valid.ne, f"Valid ne check: expected '{parameter}' != '{valid.name}=({valid.ne})', " \
                                          f"got '{parameter}' == '{valid.name}=({valid.ne})'."
        if valid.ge is not None:
            assert parameter >= valid.ge, f"Valid ge check: expected '{parameter}' >= '{valid.name}=({valid.ge})', " \
                                          f"got '{parameter}' < '{valid.name}=({valid.ge})'."
        if valid.gt is not None:
            assert parameter > valid.gt, f"Valid gt check: expected '{parameter}' > '{valid.name}=({valid.gt})', " \
                                         f"got '{parameter}' <= '{valid.name}=({valid.gt})'."
        if valid.datetime_format is not None:
            try:
                datetime.strptime(parameter, valid.datetime_format)
                result = True
            except ValueError:
                result = False
            assert result, f"Valid datetime format check: expected datetime format '{valid.datetime_format}', " \
                           f"got '{parameter}'"


__all__ = ["validate", "Valid"]
